@@ -8,15 +8,17 @@ from copy import deepcopy
 from typing import Union, Optional
 
 from .solver_methods import BaseMethod, RK45, Euler, BDF, NoDiff
-from .interface import DefaultEventHandler, Interface, _JitHelper, Model, Solution, \
-    EventHandler
+from .base import Solution
+from .interface import Interface
+from .handlers import DefaultEventHandler, EventHandler
+from .models import Model
 from .solve_states import SolveEvent, SolveStatus
 
 
 solver_methods = {'RK45': RK45, 'Euler': Euler, 'BDF': BDF, 'NoDiff': NoDiff}
 
-_Info = namedtuple('Info', ['status', 'event_id', 'step_info', 'initial_step', 'dt', 't', 'y', 'order_', 'roller',
-                           'solve_state', 'ix_eval', 'g', 'step_converged', 'event_trigger'])
+SolverInfo = namedtuple('Info', ['status', 'event_id', 'step_info', 'initial_step', 'dt', 't', 'y', 'order_', 'roller',
+                            'solve_state', 'ix_eval', 'g', 'step_converged', 'event_trigger'])
 
 try:
     FEPS = np.finfo(1.0).eps
@@ -42,20 +44,61 @@ class NumerousSolver:
 
     """
 
-    def __init__(self, model: Union[Model, object],
+    def __init__(self, model: Model,
                  use_jit: bool = False, event_handler=None,
                  **options):
 
         if not event_handler:
             event_handler = DefaultEventHandler()
         self.event_handler: EventHandler = event_handler
-        self.options = options
+
         self.use_jit = use_jit
         self.model = model
-        self.interface = model.generate_interface(use_jit)
+        self.options = {}
+        interface, time_events, external_time_events, state_events, \
+            external_state_events = model.generate_interface(use_jit)
+
+        time_events_ = [time_event._compiled_model for time_event in time_events]
+        state_events_ = [state_event._compiled_model for state_event in state_events]
+
+        self.time_events = tuple(time_events_)
+        self.state_events = tuple(state_events_)
+
+        time_events_dict = {time_event.id: time_event._compiled_model for time_event in time_events}
+        external_time_events_dict = {time_event.id: time_event for time_event in external_time_events}
+
+        state_events_dict = {state_event.id: state_event._compiled_model for state_event in state_events}
+
+        external_state_events_dict = {state_event.id: state_event for state_event in external_state_events}
+
+        self.events = dict(time_events_dict, **state_events_dict)
+        self.external_events = dict(external_time_events_dict, **external_state_events_dict)
+
+        self.interface: Interface = interface._compiled_model
+
         if not hasattr(self.interface, '_numba_type_') and use_jit:
             raise TypeError("when requesting a jitted solver, the interface must also be jitted")
 
+        self._set_options(options)
+
+        y0 = self.interface.get_states()
+        #self.g = np.array([1.0])
+        self.g = self.interface.get_state_event_results(0, y0, self.state_events)
+
+        self.info = None
+        self.y0 = y0
+
+        # Generate the solver
+        if self.use_jit:
+            self._non_compiled_solve = njit(self._generate_solver())
+            self._solve = self._compile_solver()
+        else:
+            self._solve = self._generate_solver()
+    def _set_options(self, options):
+        y0 = self.interface.get_states()
+        f0 = self.interface.get_deriv(0, y0)
+
+        self.options.update(options)
         odesolver_options = {
             'longer': self.options.get('longer', 1.2), 'shorter': self.options.get('shorter', 0.8),
             'min_step': self.options.get('min_step', 10 * FEPS), 'strict_eval': True,
@@ -64,43 +107,25 @@ class NumerousSolver:
             'submethod': self.options.get('submethod', None),
         }
         self.method_options = odesolver_options
-
-        y0 = self.interface.get_states()
-        self.g = self.interface.get_event_results(0, y0)
         method = self.options.get('method', 'RK45')
-        self.f0 = self.interface.get_deriv(0, y0)
-        if self.f0 is None:
-            logger.info("Using no states solver. If this is not intended, make sure you specify the 'get_deriv' method "
-                        "for the model")
+        if f0 is None or len(f0) == 0:
+            logger.debug("Using no derivatives solver. If this is not intended, make sure you specify the 'get_deriv' "
+                         "method for the model")
             method = "NoDiff"
+
         try:
             self.method = solver_methods[method]
         except KeyError:
             raise ValueError(f"Unknown method {self.options.get('method', 'RK45')}, "
                              f"allowed methods: {list(solver_methods.keys())}")
-
-        self.number_of_events = len(self.g)
-
-        self.interface.set_states(y0)
-
-        self.info = None
-
-        self.y0 = y0
-
         self._method = self.method(self.interface, self.use_jit, **self.method_options)
         assert issubclass(self.method, BaseMethod), f"{self.method} is not a BaseMethod"
 
-        # Generate the solver
-        if self.use_jit:
-            self._non_compiled_solve = njit(self._generate_solver())
-            self._solve = self._compile_solver()
-        else:
-            self._solve = self._generate_solver()
-
     def _generate_solver(self):
-        def _solve(interface, _solve_state, initial_step, dt_0, order, order_, roller, strict_eval,
-                   min_step, max_step, step_integrate_, g, number_of_events, step_converged, event_trigger,
-                   t0=0.0, t_end=1000.0, t_eval=np.linspace(0.0, 1000.0, 100), ix_eval=1, event_tolerance=1e-6) -> _Info:
+        def _solve(interface, time_events, state_events, _solve_state, initial_step, dt_0, order,
+                   order_, roller, strict_eval, min_step, max_step, step_integrate_, g,
+                   step_converged, event_trigger, t0=0.0, t_end=1000.0, t_eval=np.linspace(0.0, 1000.0, 100), ix_eval=1,
+                   event_tolerance=1e-6) -> SolverInfo:
 
             # Init t to t0
             imax = int(100)
@@ -110,6 +135,10 @@ class NumerousSolver:
             dt = dt_0
             interface.pre_step(t, interface.get_states())
             y = interface.get_states()
+            failed_steps = 0
+            failed_events = 0
+            warned = False
+            warned_event = False
 
             solve_status = SolveStatus.Running
             solve_event_id = SolveEvent.NoneEvent
@@ -172,8 +201,9 @@ class NumerousSolver:
             t_event = t
             y_event = y
             t_event_previous = -1
-            time_event_ixs, t_next_time_event = interface.get_next_time_event(t)
-            if time_event_ixs[0] >= 0:
+            t_next_time_event, has_time_events = interface.get_next_time_event(t, time_events)
+
+            if has_time_events:
                 te_array[2] = t_next_time_event
             else:
                 te_array[2] = max(te_array)
@@ -223,82 +253,40 @@ class NumerousSolver:
                 dt = min(dt * factor, t_end)
 
                 event_trigger = False
-                t_events = np.zeros(number_of_events) + t
-                y_events = np.zeros((len(y), number_of_events))
-
-                def sol(t, t_r, y_r):
-                    yi = np.zeros(len(y))
-                    tv = np.append(roller[1][0:order], t_r)
-                    yv = np.append(roller[2][0:order], y_r)
-                    yv = yv.reshape(order + 1, len(y)).T
-                    for i, yvi in enumerate(yv):
-                        yi[i] = np.interp(t, tv, yvi)
-                    return yi
-
-                def check_event(event_fun, ix, t_previous, y_previous, t, y):
-                    t_l = t_previous
-                    y_l = y_previous
-                    e_l = event_fun(t_l, y_l)[ix]
-                    t_r = t
-                    y_r = y
-                    e_r = event_fun(t_r, y_r)[ix]
-                    status = 0
-                    if np.sign(e_l) == np.sign(e_r):
-                        return status, t, y
-                    i = 0
-                    t_m = (t_l + t_r) / 2
-                    y_m = sol(t_m, t, y)
-
-                    while status == 0:  # bisection method
-                        e_m = event_fun(t_m, y_m)[ix]
-                        if np.sign(e_l) != np.sign(e_m):
-                            t_r = t_m
-                        elif np.sign(e_r) != np.sign(e_m):
-                            t_l = t_m
-                        if abs(e_m) < 1e-6 or abs(t_l - t_r) < 1e-6:
-                            status = 1
-                        if i > imax:
-                            status = -1
-                        t_m = (t_l + t_r) / 2
-                        y_m = sol(t_m, t, y)
-
-                    return status, t_r, sol(t_r, t, y)
+                if not step_converged:
+                    failed_steps += 1
+                    if failed_steps > 100 and not warned:
+                        print("slow convergence", dt)
+                        warned = True
 
                 if step_converged:
-                    time_event_ixs, t_next_time_event = interface.get_next_time_event(t)
-                    if time_event_ixs[0] >= 0:
+                    failed_steps = 0
+                    warned = False
+                    t_next_time_event, has_time_events = interface.get_next_time_event(t, time_events)
+                    if has_time_events:
                         te_array[2] = t_next_time_event
                     else:
                         te_array[2] = max(te_array)
 
-                    g_new = interface.get_event_results(t, y)
 
-                    up = (g <= 0) & (g_new >= 0) & (interface.get_event_directions() == 1)
-                    down = (g >= 0) & (g_new <= 0) & (interface.get_event_directions() == -1)
-                    g = g_new
-
-                    for ix in np.concatenate((np.argwhere(up), np.argwhere(down))):
-                        eps = 1.e-6  # for case to t_event = t
-                        status, t_event, y_event = check_event(interface.get_event_results, ix[0],
-                                                               t_previous, y_previous, t, y)
-                        t_events[ix[0]] = t_event - eps
-                        y_events[:, ix[0]] = y_event
-
-                if min(t_events) < t:
-                    event_trigger = True
-                    event_ix = np.argmin(t_events)
-                    t_event = t_events[event_ix]
-                    y_event = y_events[:, event_ix]
-                    g = interface.get_event_results(t_event, y_event)
+                    t_event, y_event, g, event_trigger = interface.get_state_events(g, state_events, roller, order,
+                                                                                     event_tolerance, imax,
+                                                                                     t_previous, y_previous, t, y)
 
                 if not event_trigger and step_converged:
                     y_previous = y
                     t_previous = t
-                    if abs(t - t_next_time_event) < 1e-6:
+
+                    if has_time_events and abs(t - t_next_time_event) < event_tolerance:
                         interface.set_states(y)
-                        for time_event_ix in time_event_ixs:
-                            y_previous = interface.run_time_event_action(t, y_previous, time_event_ix)  # todo: use interface to get y.
+                        
+                        external, y_previous = interface.run_events(t, y, time_events)
+
                         interface.set_states(y_previous)
+                        if external:
+                            solve_event_id = SolveEvent.TimeEvent
+                            break
+
                         solve_event_id = interface.post_time_event(t, y_previous)
                         if solve_event_id != SolveEvent.NoneEvent:
                             break
@@ -307,26 +295,44 @@ class NumerousSolver:
                     # Improve detection of event
 
                     if abs(t_event - t_event_previous) > event_tolerance:
+                        failed_events += 1
+                        if failed_events > 100 and not warned_event:
+                            print("event searching is slow. Try tightening absolute tolerance.",
+                                  abs(t_event - t_event_previous), ">", event_tolerance)
+                            warned_event = True
                         t_event_previous = t_event
                         step_converged = False  # roll back and refine search
                         dt = initial_step
-                        g = interface.get_event_results(t, y_previous)
+                        g = interface.get_state_event_results(t, y_previous, state_events)
+                        interface.unset_events(state_events)
                     else:
+                        failed_events = 0
+                        warned_event = False
                         interface.set_states(y_event)
-                        y_previous = interface.run_event_action(t_event, y_event, event_ix)
+                        external, y_previous = interface.run_events(t_event, y_event, state_events)
+                        interface.set_states(y_previous)
+                        step_converged = False
+
+                        if external:
+                            solve_event_id =  SolveEvent.StateEvent
+                            t = t_event
+                            y = y_event
+                            break
+
                         interface.set_states(y_previous)
 
                         t_previous = t_event
 
                         # TODO: Update solve in solver_methods with new solve state after changing states due to events
 
-                        solve_event_id = interface.post_event(t_event, y_previous)
+                        solve_event_id = interface.post_state_event(t_event, y_previous, "dummy")
                         # Immediate rollback in case of exit
                         t = t_previous
                         y = y_previous
 
                         if solve_event_id != SolveEvent.NoneEvent:
                             break
+                    
 
                 if step_converged:
                     interface.set_states(y)
@@ -346,7 +352,7 @@ class NumerousSolver:
                     if solve_event_id != SolveEvent.NoneEvent:
                         break
 
-            return _Info(status=solve_status, event_id=solve_event_id, step_info=step_info,
+            return SolverInfo(status=solve_status, event_id=solve_event_id, step_info=step_info,
                         dt=dt, t=t, y=np.ascontiguousarray(y), order_=order_, roller=roller, solve_state=_solve_state,
                         ix_eval=ix_eval, g=g, initial_step=initial_step, step_converged=step_converged,
                         event_trigger=event_trigger)
@@ -373,6 +379,8 @@ class NumerousSolver:
         order_ = 0
 
         args = (self.interface,
+                self.time_events,
+                self.state_events,
                 self._method.get_solver_state(),
                 initial_step,
                 initial_step,
@@ -384,7 +392,6 @@ class NumerousSolver:
                 max_step,
                 step_integrate_,
                 self.g,
-                self.number_of_events,
                 False,
                 False,
                 0.0,
@@ -415,6 +422,7 @@ class NumerousSolver:
         max_step = self.method_options.get('max_step')
         min_step = self.method_options.get('min_step')
         atol = self.method_options.get('atol')
+        rtol = self.method_options.get('rtol')
         strict_eval = self.method_options.get('strict_eval')
         step_integrate_ = self._method.step_func
         order = self._method.order
@@ -440,7 +448,9 @@ class NumerousSolver:
 
             roller = self._init_roller(order)
             order_ = 0
-            g = self.interface.get_event_results(time[0], y0)
+            g = self.interface.get_state_event_results(time[0], y0, self.state_events)
+
+            #g = self.interface.get_event_results(time[0], y0)
             step_converged = False
             event_trigger = False
         else:
@@ -455,7 +465,7 @@ class NumerousSolver:
             event_trigger = self.info.event_trigger
 
         return dt, strict_eval, step_integrate_, solve_state, roller, order_, order, initial_step, min_step, \
-               max_step, atol, g, step_converged, event_trigger
+               max_step, atol, rtol, g, step_converged, event_trigger
 
     def reset(self):
         """Resets the solution and calls the model reset method.
@@ -464,9 +474,14 @@ class NumerousSolver:
         """
         self.event_handler.reset_solution()
         self.model.reset()
+        for event in self.events.values():
+            event.reset()
+        for external_event in self.external_events.values():
+            external_event.reset()
+
         self.info = None
 
-    def solve(self, t_eval: np.array, y0: Optional[np.array] = None):
+    def solve(self, t_eval: np.array, y0: Optional[np.array] = None, **options):
         """Solves the model.
 
         :param: t_eval: Input array of all timestamps for which solution is to be evaluated
@@ -479,6 +494,9 @@ class NumerousSolver:
             self.interface.set_states(y0)
             self.info = None
 
+        if options:
+            self._set_options(options)
+
         t_eval = t_eval.astype('float')
         logger.info('Solve started')
 
@@ -504,26 +522,73 @@ class NumerousSolver:
     def _solver(self, t_eval, info=None, delta_t=None):
 
         dt, strict_eval, step_integrate_, solve_state, roller, order_, order, initial_step, min_step, max_step, \
-            atol, g, step_converged, event_trigger = self._init_solve(t_eval, info, delta_t)
+            atol, rtol, g, step_converged, event_trigger = self._init_solve(t_eval, info, delta_t)
 
         t_start = t_eval[0]
         t_end = t_eval[-1]
 
-        info = self._solve(self.interface, solve_state, initial_step, dt, order, order_,
-                           roller, strict_eval, min_step, max_step, step_integrate_, g, self.number_of_events,
-                           step_converged, event_trigger, t_start, t_end, t_eval, 1, atol*100)
+        info = self._solve(self.interface,
+                           self.time_events,
+                           self.state_events,
+                           solve_state,
+                           initial_step,
+                           dt,
+                           order,
+                           order_,
+                           roller,
+                           strict_eval,
+                           min_step,
+                           max_step,
+                           step_integrate_,
+                           g,
+                           step_converged,
+                           event_trigger,
+                           t_start,
+                           t_end,
+                           t_eval,
+                           1,
+                           atol * 1000)
 
         while info.status == SolveStatus.Running:
+            self._trigger_external_events(info.t, self.interface.get_states())
             self.event_handler.handle_solve_event(self.interface, info.event_id, info.t)
 
-            info = self._solve(self.interface, info.solve_state, initial_step, info.dt, order,
-                               info.order_, info.roller, strict_eval, min_step, max_step, step_integrate_, info.g,
-                               self.number_of_events, info.step_converged, info.event_trigger, info.t, t_end, t_eval,
-                               info.ix_eval, atol*100)
+            info = self._solve(self.interface,
+                               self.time_events,
+                               self.state_events,
+                               info.solve_state,
+                               initial_step,
+                               info.dt,
+                               order,
+                               info.order_,
+                               info.roller,
+                               strict_eval,
+                               min_step,
+                               max_step,
+                               step_integrate_,
+                               info.g,
+                               info.step_converged,
+                               info.event_trigger,
+                               info.t,
+                               t_end,
+                               t_eval,
+                               info.ix_eval,
+                               atol * 1000)
 
         self.event_handler.handle_solve_event(self.interface, info.event_id, info.t)
         return info
 
+
+    def _trigger_external_events(self, t: float, y: np.array):
+        external_events = [event for event in self.events.values() if event.is_external()]
+        for event in external_events:
+            if event.is_triggered():
+                external_event = self.external_events[event.id]
+                y = external_event.run_event_action(self.interface, t, y)
+                external_event.post_event(self.interface, t, y)
+                event.clear()
+                self.interface.set_states(y)
+
     def register_endstep(self, __end_step):
         """Unused method
 
